from __future__ import absolute_import


class KeyPair:
    """
    The interface of a key pair. A key pair is a pair consisting of a private and a public
    key used for en- and decryption.
    """

    def __init__(self, priv = None, pub = None):
        """
        Initiate a KeyPair instance using the key information provided as parameters.

        :param priv: The private key as a bytes-like object or None.
        :param pub: The public key as a bytes-like object or None.
        """

        raise NotImplementedError

    @classmethod
    def generate(cls):
        """
        :returns: A new key pair with private and public key set.
        """

        raise NotImplementedError

    @property
    def priv(self):
        """
        :returns: A bytes-like object encoding the private key of this key pair instance.
        """

        raise NotImplementedError

    @property
    def pub(self):
        """
        :returns: A bytes-like object encoding the public key of this key pair instance.
        """

        raise NotImplementedError

    def encrypt(self, data, other):
        """
        Encrypt given data using the private key stored by this KeyPair instance, for the
        public key stored by the other instance.

        :param data: The data to encrypt. A bytes-like object.
        :param other: An instance of the KeyPair class. The public key to encrypt the data
            for.
        :returns: The encrypted data.
        :raises MissingKeyException: If any key is missing to complete this operation. The
            exception message will contain (human-readable) details.
        """

        raise NotImplementedError

    def decrypt(self, data, other):
        """
        Decrypt the encrypted data using the private key stored by this KeyPair instance,
        for the public key stored by the other instance.

        :param data: The data to decrypt. A bytes-like object.
        :param other: An instance of the KeyPair class. The public key to decrypt the data
            from.
        :returns: The decrypted plain data.
        :raises MissingKeyException: If any key is missing to complete this operation. The
            exception message will contain (human-readable) details.
        """

        raise NotImplementedError

    def getSharedSecret(self, other):
        """
        Get a shared secret between the keys stored by this instance and the keys stored
        by the other instance.

        The shared secrets are generated, so that following equation is True: ::

            shared_secret(A.priv, B.pub) == shared_secret(B.priv, A.pub)

        :param other: An instance of the KeyPair class.
        :returns: The shared secret, as a bytes-like object.
        :raises MissingKeyException: If any key is missing to complete this operation. The
            exception message will contain (human-readable) details.
        """

        raise NotImplementedError


from .utils import keyexchange, reduce_path, create_leaf_node

# functions on the receiver end

def process_group_message(client, server, initiator_name, setup_key_pub, copath):
    # construct shared secret
    initiator = server.getBundle(initiator_name)
    leaf_key = keyexchange(client.pre_key.priv, client.iden_key.priv, initiator.iden_key.pub, setup_key_pub)
    secret = create_leaf_node(leaf_key)
    recon = reduce_path(secret, copath)
    return recon.priv


import math
from pdb import set_trace
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout
import matplotlib.pyplot as plt

class Node:

    def __init__(self, priv=None, pub=None, name=None) -> None:
        self.parent = None
        self.left = None
        self.right = None
        self.priv = priv
        self.pub = pub
        self.name = name
    
    def set_parent(self, node):
        self.parent = node
    
    def is_root(self):
        return self.parent == None
    
    def is_leaf(self):
        return self.left == None and self.right == None

    def __repr__(self) -> str:
        return f"Pub {self.pub}"
    
    def __str__(self) -> str:
        return f"Pub {self.pub}"
    
class PublicNode(Node):

    def __init__(self, priv, pub) -> None:
        self.priv = priv
        self.pub = pub
        super(PublicNode,self).__init__()

class SecretNode(Node):

    def __init__(self, priv, pub) -> None:
        self.priv = priv
        self.pub = pub
        super(SecretNode,self).__init__()

class ProofNode(Node):
    def __init__(self, pub, name=None) -> None:
        super().__init__(None, pub, name)
    
    def __repr__(self) -> str:
        return f"Name {self.name}"
    
    def __str__(self) -> str:
        return f"Name{self.name}"

def nleft(x):
    return int(math.pow(2, math.ceil(math.log2(x))-1))

def get_sibling(node):
    if node.is_root():
        return None
    parent, is_left = node.parent
    return parent.right if is_left else parent.left

def get_leaves(node):
    if node.is_leaf():
        return [node]
    return get_leaves(node.left) + get_leaves(node.right)

def to_networkx(node: Node):
    def get_edges(node):
        if node.is_leaf():
            return []
        return get_edges(node.left) + get_edges(node.right) + [(node.pub, node.left.pub), (node.pub, node.right.pub)]
    
    def get_nodes(node):
        if node.is_leaf():
            return [node]
        return get_nodes(node.left) + [node] + get_nodes(node.right)

    gr = nx.DiGraph()
    for n in get_nodes(node):
        gr.add_node(n.pub, name=n.name if n.name is not None else "")
    edges = get_edges(node)
    gr.add_edges_from(edges)
    return gr

def plot(G, title, fname):
    # same layout using matplotlib with no labels
    plt.title(f'{title}')
    pos=graphviz_layout(G, prog='dot')
    labels = nx.get_node_attributes(G, 'name') 
    nx.draw(G, pos, labels=labels, node_shape='s', node_size=1000, arrows=True)
    plt.savefig(f'{fname}.png')


from pyxart.keys import KeyPairCurve25519
class Client:

    def __init__(self, name) -> None:
        self.name = name
        self.iden_key = KeyPairCurve25519.generate()
        self.pre_key = KeyPairCurve25519.generate()
    
    def __repr__(self) -> str:
        return f"Client{self.name}"
    
    def __str__(self) -> str:
        return f"Client name is {self.name}"


from pyxart.client import Client
from collections import namedtuple

Bundle = namedtuple('Bundle', 'iden_key pre_key')
class Server:

    def __init__(self) -> None:
        self.clients = {}
    
    def register(self, client):
        self.clients[client.name] = client
    
    def getBundle(self, client_name):
        client = self.clients[client_name]
        return Bundle(client.iden_key, client.pre_key)


from turtle import pu
from itertools import tee
import pyxart.keys as keys
from nacl.bindings.crypto_scalarmult import crypto_scalarmult
from collections import namedtuple
from .tree import ProofNode, nleft, get_sibling, SecretNode, PublicNode, Node
from .utils import keyexchange, get_pub, dh, reduce_path, create_leaf_node
from collections import namedtuple


GroupSetupMessage = namedtuple('GroupSetupMessage','initiator participants setup_key tree')

def compute_parent(left, right):
    return dh(left.priv, right.pub)

def compute_tree_secret(secrets):
    if len(secrets) == 1:
        return secrets[0]
    num_nodes_in_left = nleft(len(secrets))
    left = compute_tree_secret(secrets[:num_nodes_in_left])
    right = compute_tree_secret(secrets[num_nodes_in_left:])
    # compute parent node
    parent = compute_parent(left, right)
    parent.left = left
    parent.right = right
    left.parent = (parent, True)
    right.parent = (parent, False)
    return parent

def create_proof_node(node):
    return ProofNode(node.pub)


def create_group(clients, server):
    secrets = []
    setup_key = keys.KeyPairCurve25519.generate()
    creator_leaf_key = keys.KeyPairCurve25519.generate()
    secrets.append(create_leaf_node(priv=creator_leaf_key.priv, name=f"Group creator's ({clients[0].name})initiation key"))
    creator = server.getBundle(clients[0].name)
    for participant in clients[1:]:
        bundle = server.getBundle(participant.name)
        leaf_key = keyexchange(setup_key.priv, creator.iden_key.priv, bundle.iden_key.pub, bundle.pre_key.pub)
        secrets.append(create_leaf_node(priv=leaf_key, name=f"Shared key between ({clients[0].name}, {participant.name})"))
    tree = compute_tree_secret(secrets)
    return create_setup_message(tree, clients, setup_key), tree.priv

def create_copath(leaf_node):
    """
    Return public keys on the copath for index^{th} leaf
    """
    while not leaf_node.is_root():
        yield create_proof_node(get_sibling(leaf_node))
        leaf_node, _ = leaf_node.parent

def create_proof_tree(tree: Node) -> ProofNode:
    if tree is None:
        return tree
    proof_root = ProofNode(tree.pub, tree.name)
    proof_root.left = create_proof_tree(tree.left)
    proof_root.right = create_proof_tree(tree.right)
    if proof_root.left is not None:
        proof_root.left.parent = (proof_root, True)
    if proof_root.right is not None:
        proof_root.right.parent = (proof_root, False)
    return proof_root

def create_setup_message(tree, clients, setup_key):
    return GroupSetupMessage(clients[0].name, [p.name for p in clients[1:]], setup_key.pub, create_proof_tree(tree))


from nacl.bindings.crypto_scalarmult import crypto_scalarmult
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from .tree import ProofNode, nleft, get_sibling, SecretNode, PublicNode, Node
from cryptography.hazmat.backends import default_backend
import libnacl

def keyexchange(s1, s2, o1, o2):
    return crypto_scalarmult(s1, o2)

def reduce_path(secret, path):
    for node in path:
        secret = dh(secret.priv, node.pub)
    return secret


def kdf(secret_key_material):
    """
    :param secret_key_material: A bytes-like object encoding the secret key material.
    :returns: A bytes-like object encoding the shared secret key.
    """

    salt = b"\x00"
    input_key_material = secret_key_material
    hkdf = HKDF(algorithm=hashes.SHA256(), length=32, salt=None, info=None, backend=default_backend())
    return hkdf.derive(input_key_material)

def get_pub(priv):
    return libnacl.crypto_scalarmult_base(priv)

def dh(priv, pub):
    shared_secret = crypto_scalarmult(priv, pub)
    parent_priv = kdf(shared_secret)
    parent_pub = get_pub(parent_priv)
    return Node(parent_priv, parent_pub)

def create_leaf_node(priv, name=None):
    return Node(priv, get_pub(priv), name)


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyxart',
 'pyxart.client',
 'pyxart.exceptions',
 'pyxart.group',
 'pyxart.keys',
 'pyxart.server']

package_data = \
{'': ['*']}

install_requires = \
['XEdDSA>=0.6.0,<0.7.0',
 'cryptography>=37.0.4,<38.0.0',
 'matplotlib>=3.5.3,<4.0.0',
 'networkx>=2.8.6,<3.0.0',
 'pynacl>=1.5.0,<2.0.0']

setup_kwargs = {
    'name': 'pyxart',
    'version': '0.1.0',
    'description': 'Python implementation of Asynchronous Ratchet Trees',
    'long_description': '# pyxart\nPython implementation of Asynchronous Ratchet Trees\n\n# references\n\n- https://research.facebook.com/publications/on-ends-to-ends-encryption-asynchronous-group-messaging-with-strong-security-guarantees/',
    'author': 'Pranay Anchuri',
    'author_email': 'pranay@xmtp.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
